#include "pcheader.h"

// FUSE: Fallo's Ultima Online Server Emulator
// Internet-based, multiplayer game server engine
// Copyright (C) 1998 Erik McClenney, aka Fallo
// email: fallo@stratics.com

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// or visit their website at http://www.gnu.org/

#include "server.h"

// add verdata.mul processing for landtiles, map, staticidx, static, multiidx
// already have for arttiles, multi

// if MAPCACHESIZE < 1, we are in trouble
#define MAPCACHESIZE (50)
#define STATICCACHESIZE (50)
#define TILECACHESIZE (50)
#define LANDTILECACHESIZE (50)

#define IDXBLOCKSIZE 12

#pragma pack(push, 1)
struct IdxBlock {
	int pos;
	int size;
	unsigned flags;
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct MapPos {
	unsigned short tile;
	unsigned char z;
} ;
#pragma pack(pop)

#pragma pack(push, 1)
struct MapBlock {
	// what the heck is this?
	unsigned val;
	// [y][x]
	MapPos pos[MAPBLOCKDIM][MAPBLOCKDIM];
} ;
#pragma pack(pop)

struct cacheitem {
	unsigned key;
	MapBlock item;
	cacheitem *next;
} ;

struct staticcacheitem {
	unsigned key;
	staticcacheitem *next;
	StaticDescriptor *items;
} ;

#pragma pack(push, 1)
struct TileBlock {
	// what the heck is this?
	unsigned unk;
	TileData tiles[32];
} ;
#pragma pack(pop)

struct tilecacheitem {
	unsigned key;
	TileBlock tb;
	tilecacheitem *next;
} ;

#pragma pack(push, 1)
struct LandTileBlock {
	// what the heck is this?
	unsigned unk;
	LandTileData tiles[32];
} ;
#pragma pack(pop)

struct landtilecacheitem {
	unsigned key;
	LandTileBlock tb;
	landtilecacheitem *next;
} ;

#define VERDATA_MULTI 0x0e
#define VERDATA_TILES 0x1e

#pragma pack(push, 1)
struct VerDataIdx {
  int type;
  int blockNum;
  int pos;
  int size;
  int serial;
};
#pragma pack(pop)

struct VerDataBlock {
	unsigned int count;
	VerDataIdx idx[1];
} ;

char *verdata = NULL;
char *multimul = NULL, *multiidx = NULL;

FILE *mapfile, *tilefile, *staticidx, *staticfile;

// this critical section is for the map cache
// use mapcsect for all items which are atomically accessed
// for now, map data and tile data are atomically accessed
// multi data, and verdata are completely loaded into ram
// static data requires a separate critical section
CRITICAL_SECTION mapcsect, staticcsect;

cacheitem mapcachelist[MAPCACHESIZE];
cacheitem *mapfreelist, *mapcache;

staticcacheitem staticcachelist[STATICCACHESIZE];
staticcacheitem *staticfreelist, *staticcache;

tilecacheitem tilecachelist[TILECACHESIZE];
tilecacheitem *tilefreelist, *tilecache;

landtilecacheitem landtilecachelist[LANDTILECACHESIZE];
landtilecacheitem *landtilefreelist, *landtilecache;

// most recently used item goes on the *bottom* of the list
// also keep a sorted list of items for a binary search

char GetMapHeight(unsigned short, unsigned short);
void PrepareMultiStuff();
void UnprepareMultiStuff();

int usesverdata = 0;

int UsesVerdata(void)
{
	return usesverdata;
}

bool MapSetup(const char *mapfilename, const char *tilefilename, const char *staidxfilename, const char *staticsfilename, const char *verfilename, const char *multiidxfilename, const char *multifilename)
{
	int i;
	int i1 = sizeof(MapBlock), i2 = MAPBLOCK;
	long size;
	FILE *ifile;

	mapfile = fopen(mapfilename, "rb");
	if(mapfile == NULL)
		return false;

	tilefile = fopen(tilefilename, "rb");
	if(tilefile == NULL)
		return false;

	staticidx = fopen(staidxfilename, "rb");
	if(staticidx == NULL)
		return false;

	staticfile = fopen(staticsfilename, "rb");
	if(staticfile == NULL)
		return false;

	InitializeCriticalSection(&mapcsect);
	InitializeCriticalSection(&staticcsect);

	for(i = 0; i < MAPCACHESIZE; i++) {
		if(i + 1 < MAPCACHESIZE)
			mapcachelist[i].next = mapcachelist + i + 1;
		else
			mapcachelist[i].next = NULL;
	}

	mapfreelist = mapcachelist;
	mapcache = NULL;

	for(i = 0; i < STATICCACHESIZE; i++) {
		if(i + 1 < STATICCACHESIZE)
			staticcachelist[i].next = staticcachelist + i + 1;
		else
			staticcachelist[i].next = NULL;

		staticcachelist[i].items = NULL;
	}

	staticfreelist = staticcachelist;
	staticcache = NULL;

	for(i = 0; i < TILECACHESIZE; i++) {
		if(i + 1 < TILECACHESIZE)
			tilecachelist[i].next = tilecachelist + i + 1;
		else
			tilecachelist[i].next = NULL;
	}

	tilefreelist = tilecachelist;
	tilecache = NULL;

	for(i = 0; i < LANDTILECACHESIZE; i++) {
		if(i + 1 < LANDTILECACHESIZE)
			landtilecachelist[i].next = landtilecachelist + i + 1;
		else
			landtilecachelist[i].next = NULL;
	}

	landtilefreelist = landtilecachelist;
	landtilecache = NULL;

	ifile = fopen(verfilename, "rb");

	if(ifile == NULL) {
		usesverdata = 0;
	}
	else {
		fseek(ifile, 0, SEEK_END);
		size = ftell(ifile);

		verdata = (char *)malloc(size);

		fseek(ifile, 0, SEEK_SET);
		fread(verdata, size, 1, ifile);
		fclose(ifile);

		usesverdata = 1;
	}

	ifile = fopen(multiidxfilename, "rb");
	if(ifile == NULL)
		return false;

	fseek(ifile, 0, SEEK_END);
	size = ftell(ifile);

	multiidx = (char *)malloc(size);

	fseek(ifile, 0, SEEK_SET);
	fread(multiidx, size, 1, ifile);
	fclose(ifile);

	ifile = fopen(multifilename, "rb");
	if(ifile == NULL)
		return false;

	fseek(ifile, 0, SEEK_END);
	size = ftell(ifile);

	multimul = (char *)malloc(size);

	fseek(ifile, 0, SEEK_SET);
	fread(multimul, size, 1, ifile);
	fclose(ifile);

	// prepare multi stuff
	PrepareMultiStuff();

	return true;
}

void MapRelease(void)
{
	int i;

	DeleteCriticalSection(&mapcsect);
	DeleteCriticalSection(&staticcsect);

	for(i = 0; i < STATICCACHESIZE; i++) {
		if(staticcachelist[i].items != NULL)
			free(staticcachelist[i].items);
	}

	if(verdata != NULL)
		free(verdata);

	if(multiidx != NULL)
		free(multiidx);

	if(multimul != NULL)
		free(multimul);

	// unprepare multi stuff
	UnprepareMultiStuff();

	fclose(mapfile);
	fclose(tilefile);
	fclose(staticidx);
	fclose(staticfile);
}

const MapBlock *GetCachedMapBlock(int x, int y)
{
	unsigned key;
	cacheitem *list, *prev;
	bool set = false;
	int pos;

	key = (x << 16) | y;

	if(mapcache != NULL) {
		prev = NULL;
		list = mapcache;

		while(list -> next != NULL) {
			if(list -> key == key) {
				break;
			}
			else {
				prev = list;
				list = list -> next;
			}
		}

		// this way, we get to check the key
		// if we are at the end of the list
		// and if we already know it matches
		if(list -> key == key) {
			// put list at the top of the list
			// if prev is NULL, we are at the top
			if(prev != NULL) {
				prev -> next = list -> next;
				list -> next = mapcache;
				mapcache = list;
			}
		}
		// didn't find it
		else {
			// we have more room to add to the cache list
			if(mapfreelist != NULL) {
				list = mapfreelist;
				mapfreelist = mapfreelist -> next;
				list -> next = mapcache;
				mapcache = list;
			}
			// we don't have more room for the cache, use
			// the current item
			else {
				if(prev != NULL) {
					prev -> next = list -> next;
					list -> next = mapcache;
					mapcache = list;
				}
			}

			// setup new list item
			set = true;
		}
	}
	else {
		if(mapfreelist != NULL) {
			list = mapfreelist;
			mapfreelist = mapfreelist -> next;
			list -> next = mapcache;
			mapcache = list;

			// set new list item
			set = true;
		}
		// try setting a non-zero mapcachesize!!
		else {
			list = NULL;
		}
	}

	if(set) {
		pos = x * MAPYSTRIP + y * MAPBLOCK/* + MAPBLOCKHDR + ymod * MAPBLOCKDIM * MAPPOSBYTES + xmod * MAPPOSBYTES + 2*/;
		fseek(mapfile, pos, SEEK_SET);
		fread(&(list -> item), MAPBLOCK, 1, mapfile);
		list -> key = key;
	}

	return &(list -> item);
}

char GetMapHeight(unsigned short ix, unsigned short iy)
{
	int x = ix, y = iy, xdiv, ydiv, xmod, ymod;
	const MapBlock *mb;
	unsigned char z;

	if(ix >= MAPX || iy >= MAPY) {
		z = 0;
	}
	else {
		xdiv = x / MAPBLOCKDIM;
		xmod = x % MAPBLOCKDIM;
		ydiv = y / MAPBLOCKDIM;
		ymod = y % MAPBLOCKDIM;

		EnterCriticalSection(&mapcsect);

		mb = GetCachedMapBlock(xdiv, ydiv);

		z = mb -> pos[ymod][xmod].z;

		LeaveCriticalSection(&mapcsect);
	}

	return z;
}

// new average method. average z's on one of the diagonals. to select which, determine
// the diagonal with the smallest delta z. question: how do we resolve conflicts? do we
// select the diagonal with the smallest/largest sum, or do we favor one of the diagonals?
FUSELIBFNDEF char GetMapZ(unsigned short ix, unsigned short iy)
{
	short zdif1, zdif2;

	zdif1 = (char)GetMapHeight(ix, iy);
	zdif1 -= (char)GetMapHeight(ix + 1, iy + 1);

	if(zdif1 < 0)
		zdif1 = -zdif1;

	zdif2 = (char)GetMapHeight(ix + 1, iy);
	zdif2 -= (char)GetMapHeight(ix, iy + 1);

	if(zdif2 < 0)
		zdif2 = -zdif2;

	if(zdif1 < zdif2) {
		zdif1 = (char)GetMapHeight(ix, iy);
		zdif1 += (char)GetMapHeight(ix + 1, iy + 1);
		zdif1 /= 2;
	}
	else {
		zdif1 = (char)GetMapHeight(ix + 1, iy);
		zdif1 += (char)GetMapHeight(ix, iy + 1);
		zdif1 /= 2;
	}

	return (char)zdif1;
}

FUSELIBFNDEF unsigned short GetMapTile(unsigned short ix, unsigned short iy)
{
	int x = ix, y = iy, xdiv, ydiv, xmod, ymod;
	const MapBlock *mb;
	unsigned short tile;

	xdiv = x / MAPBLOCKDIM;
	xmod = x % MAPBLOCKDIM;
	ydiv = y / MAPBLOCKDIM;
	ymod = y % MAPBLOCKDIM;

	EnterCriticalSection(&mapcsect);

	mb = GetCachedMapBlock(xdiv, ydiv);

	tile = mb -> pos[ymod][xmod].tile;

	LeaveCriticalSection(&mapcsect);

	return tile;
}

const TileBlock *GetCachedTileBlock(unsigned short model)
{
	unsigned key;
	tilecacheitem *list, *prev;
	bool set = false;
	int pos, i;
	VerDataBlock *vdb;
	TileBlock *tb;

	key = model;

	if(tilecache != NULL) {
		prev = NULL;
		list = tilecache;

		while(list -> next != NULL) {
			if(list -> key == key) {
				break;
			}
			else {
				prev = list;
				list = list -> next;
			}
		}

		// this way, we get to check the key
		// if we are at the end of the list
		// and if we already know it matches
		if(list -> key == key) {
			// put list at the top of the list
			// if prev is NULL, we are at the top
			if(prev != NULL) {
				prev -> next = list -> next;
				list -> next = tilecache;
				tilecache = list;
			}
		}
		// didn't find it
		else {
			// we have more room to add to the cache list
			if(tilefreelist != NULL) {
				list = tilefreelist;
				tilefreelist = tilefreelist -> next;
				list -> next = tilecache;
				tilecache = list;
			}
			// we don't have more room for the cache, use
			// the current item
			else {
				if(prev != NULL) {
					prev -> next = list -> next;
					list -> next = tilecache;
					tilecache = list;
				}
			}

			// setup new list item
			set = true;
		}
	}
	else {
		if(tilefreelist != NULL) {
			list = tilefreelist;
			tilefreelist = tilefreelist -> next;
			list -> next = tilecache;
			tilecache = list;

			// set new list item
			set = true;
		}
		// try setting a non-zero mapcachesize!!
		else {
			list = NULL;
		}
	}

	if(set) {
		vdb = (VerDataBlock *)verdata;
		tb = NULL;

		if(UsesVerdata()) {
			for(i = 0; i < (int)vdb -> count; i++) {
				if(vdb -> idx[i].type == VERDATA_TILES) {
					// must add 0x200 to get past land tiles
					if((int)(key + 0x200) == vdb -> idx[i].blockNum) {
						tb = (TileBlock *)(verdata + vdb -> idx[i].pos);
						break;
					}
				}
			}
		}

		if(tb == NULL) {
			pos = (sizeof(LandTileBlock) << 9) + sizeof(TileBlock) * key;
			fseek(tilefile, pos, SEEK_SET);
			fread(&(list -> tb), sizeof(TileBlock), 1, tilefile);
			list -> key = key;
		}
		else {
			memcpy(&(list -> tb), tb, sizeof(TileBlock));
			list -> key = key;
		}
	}

	return &(list -> tb);
}

FUSELIBFNDEF void GetTileName(unsigned short itemid, char *name)
{
	const TileBlock *tb;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedTileBlock(itemid >> 5);
	strcpy(name, tb -> tiles[itemid & 0x1f].name);

	LeaveCriticalSection(&mapcsect);
}

FUSELIBFNDEF unsigned GetTileFlags(unsigned short itemid)
{
	const TileBlock *tb;
	unsigned flags;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedTileBlock(itemid >> 5);
	flags = htonl(tb -> tiles[itemid & 0x1f].flags);

	LeaveCriticalSection(&mapcsect);

	return flags;
}

FUSELIBFNDEF unsigned char GetTileWeight(unsigned short itemid)
{
	const TileBlock *tb;
	unsigned char data;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedTileBlock(itemid >> 5);
	data = tb -> tiles[itemid & 0x1f].data[TILE_WEIGHT];

	LeaveCriticalSection(&mapcsect);

	return data;
}

FUSELIBFNDEF unsigned char GetTileEquip(unsigned short itemid)
{
	const TileBlock *tb;
	unsigned char data;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedTileBlock(itemid >> 5);
	data = tb -> tiles[itemid & 0x1f].data[TILE_EQUIP];

	LeaveCriticalSection(&mapcsect);

	return data;
}

FUSELIBFNDEF unsigned char GetTileHeight(unsigned short itemid)
{
	const TileBlock *tb;
	unsigned char data;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedTileBlock(itemid >> 5);
	data = tb -> tiles[itemid & 0x1f].data[TILE_HEIGHT];

	LeaveCriticalSection(&mapcsect);

	return data;
}

FUSELIBFNDEF void GetTileData(unsigned short itemid, TileData *td)
{
	const TileBlock *tb;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedTileBlock(itemid >> 5);
	memcpy(td, tb -> tiles + (itemid & 0x1f), sizeof(TileData));
	td -> flags = htonl(td -> flags);

	LeaveCriticalSection(&mapcsect);
}

const LandTileBlock *GetCachedLandTileBlock(unsigned short model)
{
	unsigned key;
	landtilecacheitem *list, *prev;
	bool set = false;
	int pos, i;
	VerDataBlock *vdb;
	LandTileBlock *tb;

	key = model;

	if(landtilecache != NULL) {
		prev = NULL;
		list = landtilecache;

		while(list -> next != NULL) {
			if(list -> key == key) {
				break;
			}
			else {
				prev = list;
				list = list -> next;
			}
		}

		// this way, we get to check the key
		// if we are at the end of the list
		// and if we already know it matches
		if(list -> key == key) {
			// put list at the top of the list
			// if prev is NULL, we are at the top
			if(prev != NULL) {
				prev -> next = list -> next;
				list -> next = landtilecache;
				landtilecache = list;
			}
		}
		// didn't find it
		else {
			// we have more room to add to the cache list
			if(landtilefreelist != NULL) {
				list = landtilefreelist;
				landtilefreelist = landtilefreelist -> next;
				list -> next = landtilecache;
				landtilecache = list;
			}
			// we don't have more room for the cache, use
			// the current item
			else {
				if(prev != NULL) {
					prev -> next = list -> next;
					list -> next = landtilecache;
					landtilecache = list;
				}
			}

			// setup new list item
			set = true;
		}
	}
	else {
		if(landtilefreelist != NULL) {
			list = landtilefreelist;
			landtilefreelist = landtilefreelist -> next;
			list -> next = landtilecache;
			landtilecache = list;

			// set new list item
			set = true;
		}
		// try setting a non-zero mapcachesize!!
		else {
			list = NULL;
		}
	}

	if(set) {
		vdb = (VerDataBlock *)verdata;
		tb = NULL;

		if(UsesVerdata()) {
			for(i = 0; i < (int)vdb -> count; i++) {
				if(vdb -> idx[i].type == VERDATA_TILES) {
					// land tiles start at 0
					if((int)(key) == vdb -> idx[i].blockNum) {
						tb = (LandTileBlock *)(verdata + vdb -> idx[i].pos);
						break;
					}
				}
			}
		}

		if(tb == NULL) {
			pos = /*(sizeof(LandTileBlock) << 9) + */sizeof(LandTileBlock) * key;
			fseek(tilefile, pos, SEEK_SET);
			fread(&(list -> tb), sizeof(LandTileBlock), 1, tilefile);
			list -> key = key;
		}
		else {
			memcpy(&(list -> tb), tb, sizeof(LandTileBlock));
			list -> key = key;
		}
	}

	return &(list -> tb);
}

FUSELIBFNDEF void GetLandTileName(unsigned short itemid, char *name)
{
	const LandTileBlock *tb;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedLandTileBlock(itemid >> 5);
	strcpy(name, tb -> tiles[itemid & 0x1f].name);

	LeaveCriticalSection(&mapcsect);
}

FUSELIBFNDEF unsigned GetLandTileFlags(unsigned short itemid)
{
	const LandTileBlock *tb;
	unsigned flags;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedLandTileBlock(itemid >> 5);
	flags = htonl(tb -> tiles[itemid & 0x1f].flags);

	LeaveCriticalSection(&mapcsect);

	return flags;
}

FUSELIBFNDEF void GetLandTileData(unsigned short itemid, LandTileData *td)
{
	const LandTileBlock *tb;

	if(itemid >= 0x4000)
		itemid = 0;

	EnterCriticalSection(&mapcsect);

	tb = GetCachedLandTileBlock(itemid >> 5);
	memcpy(td, tb -> tiles + (itemid & 0x1f), sizeof(LandTileData));
	td -> flags = htonl(td -> flags);

	LeaveCriticalSection(&mapcsect);
}

#define STATICITEMSIZE 7

#pragma pack(push, 1)
struct StaticItem {
	unsigned short model;
	unsigned char x;
	unsigned char y;
	unsigned char z;
	unsigned short color;
} ;
#pragma pack(pop)

const StaticDescriptor *GetCachedStaticBlock(int x, int y)
{
	unsigned key;
	staticcacheitem *list, *prev;
	bool set = false;
	int pos, i, xbase, ybase, count;
	StaticItem *items;
	IdxBlock idx;

	key = (x << 16) | y;

	if(staticcache != NULL) {
		prev = NULL;
		list = staticcache;

		while(list -> next != NULL) {
			if(list -> key == key) {
				break;
			}
			else {
				prev = list;
				list = list -> next;
			}
		}

		// this way, we get to check the key
		// if we are at the end of the list
		// and if we already know it matches
		if(list -> key == key) {
			// put list at the top of the list
			// if prev is NULL, we are at the top
			if(prev != NULL) {
				prev -> next = list -> next;
				list -> next = staticcache;
				staticcache = list;
			}
		}
		// didn't find it
		else {
			// we have more room to add to the cache list
			if(staticfreelist != NULL) {
				list = staticfreelist;
				staticfreelist = staticfreelist -> next;
				list -> next = staticcache;
				staticcache = list;
			}
			// we don't have more room for the cache, use
			// the current item
			else {
				if(prev != NULL) {
					prev -> next = list -> next;
					list -> next = staticcache;
					staticcache = list;
				}
			}

			// setup new list item
			set = true;
		}
	}
	else {
		if(staticfreelist != NULL) {
			list = staticfreelist;
			staticfreelist = staticfreelist -> next;
			list -> next = staticcache;
			staticcache = list;

			// set new list item
			set = true;
		}
		// try setting a non-zero mapcachesize!!
		else {
			list = NULL;
		}
	}

	if(set) {
		list -> key = key;

		if(list -> items != NULL)
			free(list -> items);

		pos = x * ((MAPY / MAPBLOCKDIM) * IDXBLOCKSIZE) + y * IDXBLOCKSIZE;
		fseek(staticidx, pos, SEEK_SET);
		fread(&idx, sizeof(IdxBlock), 1, staticidx);

		if(idx.pos == -1) {
			count = 0;
			list -> items = (StaticDescriptor *)malloc(sizeof(StaticDescriptor));
			list -> items -> count = count;
		}
		else {
			count = idx.size / STATICITEMSIZE;
			list -> items = (StaticDescriptor *)malloc(sizeof(StaticDescriptor) + sizeof(StaticBlock) * (count - 1));
			list -> items -> count = count;
			items = (StaticItem *)malloc(idx.size);

			fseek(staticfile, idx.pos, SEEK_SET);
			fread(items, idx.size, 1, staticfile);

			xbase = x * MAPBLOCKDIM;
			ybase = y * MAPBLOCKDIM;

			for(i = 0; i < count; i++) {
				list -> items -> items[i].model = items[i].model;
				list -> items -> items[i].x = xbase + items[i].x;
				list -> items -> items[i].y = ybase + items[i].y;
				list -> items -> items[i].z = items[i].z;
				list -> items -> items[i].color = items[i].color;
			}

			free(items);
		}
	}

	return list -> items;
}

/*FUSELIBFNDEF int GetStaticBlockSize(const posinfo &);
FUSELIBFNDEF int GetStaticItem(const posinfo &, int, unsigned short &, unsigned short &, unsigned short &, unsigned char &, unsigned short &);*/
/*FUSELIBFNDEF int GetStaticBlockSize(const posinfo &p)
{
	int pos;
	int xdiv, ydiv;
	IdxBlock idx;

	xdiv = p.x / MAPBLOCKDIM;
	ydiv = p.y / MAPBLOCKDIM;

	pos = xdiv * ((MAPY / MAPBLOCKDIM) * IDXBLOCKSIZE) + ydiv * IDXBLOCKSIZE;
	//pos = ydiv * ((MAPX / MAPBLOCKDIM) * IDXBLOCKSIZE) + xdiv * IDXBLOCKSIZE;
	fseek(staticidx, pos, SEEK_SET);
	fread(&idx, sizeof(IdxBlock), 1, staticidx);

	if(idx.pos == -1) {
		return 0;
	}
	else {
		return idx.size / STATICITEMSIZE;
	}
}

FUSELIBFNDEF int GetStaticItem(const posinfo &p, int index, unsigned short &model, unsigned short &x, unsigned short &y, unsigned char &z, unsigned short &color)
{
	int pos;
	int xdiv, ydiv;
	IdxBlock idx;
	StaticItem sitem;

	xdiv = p.x / MAPBLOCKDIM;
	ydiv = p.y / MAPBLOCKDIM;

	pos = xdiv * ((MAPY / MAPBLOCKDIM) * IDXBLOCKSIZE) + ydiv * IDXBLOCKSIZE;
	//pos = ydiv * ((MAPX / MAPBLOCKDIM) * IDXBLOCKSIZE) + xdiv * IDXBLOCKSIZE;
	fseek(staticidx, pos, SEEK_SET);
	fread(&idx, sizeof(IdxBlock), 1, staticidx);

	if(idx.pos == -1) {
		return 0;
	}
	else if(index >= 0 && index < idx.size / STATICITEMSIZE) {
		pos = idx.pos + index * STATICITEMSIZE;
		fseek(staticfile, pos, SEEK_SET);
		fread(&sitem, sizeof(StaticItem), 1, staticfile);

		model = sitem.model;
		x = xdiv * MAPBLOCKDIM + sitem.x;
		y = ydiv * MAPBLOCKDIM + sitem.y;
		z = sitem.z;
		color = sitem.color;

		return 1;
	}
	else {
		return 0;
	}
}*/

FUSELIBFNDEF void LockStatics(void)
{
	EnterCriticalSection(&staticcsect);
}

FUSELIBFNDEF void UnlockStatics(void)
{
	LeaveCriticalSection(&staticcsect);
}

FUSELIBFNDEF const StaticDescriptor *GetStaticDescriptor(const posinfo &p)
{
	int xdiv, ydiv;

	// ideally, make sure that we are in the static critical section, else return NULL

	xdiv = p.x / MAPBLOCKDIM;
	ydiv = p.y / MAPBLOCKDIM;

	return GetCachedStaticBlock(xdiv, ydiv);
}

FUSELIBFNDEF void LockMultis(void)
{
	//EnterCriticalSection(&multicsect);
}

FUSELIBFNDEF void UnlockMultis(void)
{
	//LeaveCriticalSection(&multicsect);
}

MultiDescriptor mdnull;
MultiDescriptor *md[0x1000];

void PrepareMultiStuff(void)
{
	unsigned i,  count;
	VerDataBlock *vdb;
	IdxBlock *idx;

	mdnull.count = 0;

	for(i = 0; i < 0x1000; i++)
		md[i] = NULL;

	vdb = (VerDataBlock *)verdata;

	if(UsesVerdata()) {
		for(i = 0; i < vdb -> count; i++) {
			if(vdb -> idx[i].type == VERDATA_MULTI) {
				count = vdb -> idx[i].size / sizeof(MultiData);
				md[vdb -> idx[i].blockNum] = (MultiDescriptor *)malloc(sizeof(MultiDescriptor) + sizeof(MultiData) * (count - 1));
				md[vdb -> idx[i].blockNum] -> count = count;
				memcpy(md[vdb -> idx[i].blockNum] -> items, verdata + vdb -> idx[i].pos, vdb -> idx[i].size);
			}
		}
	}

	idx = (IdxBlock *)multiidx;

	for(i = 0; i < 0x1000; i++) {
		if(md[i] == NULL) {
			if(idx[i].pos == -1) {
				count = 0;
				md[i] = (MultiDescriptor *)malloc(sizeof(MultiDescriptor));
				md[i] -> count = count;
			}
			else {
				count = idx[i].size / sizeof(MultiData);
				md[i] = (MultiDescriptor *)malloc(sizeof(MultiDescriptor) + sizeof(MultiData) * (count - 1));
				md[i] -> count = count;
				memcpy(md[i] -> items, multimul + idx[i].pos, idx[i].size);
			}
		}
	}

	// test crap; test for possible values of 'flag'
	/*int j, flag, zeros = 0, ones = 0, others = 0;

	for(i = 0; i < 0x1000; i++) {
		for(j = 0; j < md[i] -> count; j++) {
			flag = md[i] -> items[j].flag;

			if(flag == 0)
				zeros++;
			else if(flag == 1)
				ones++;
			else
				others++;
		}
	}*/
}

void UnprepareMultiStuff(void)
{
	unsigned i;

	for(i = 0; i < 0x1000; i++)
		free(md[i]);
}

FUSELIBFNDEF const MultiDescriptor *GetMultiDescriptor(unsigned short model)
{
	unsigned short key;

	key = model - 0x4000;

	if(key < 0x1000)
		return md[key];
	else
		return &mdnull;
}
